(* =============================================================================
================ FONCTIONS GENERALES SUR LES LISTES ============================
===============================================================================*)



let suppression p l =  
  (* suppression des lments de la liste qui vrifie la proprit p *)
  List.rev (List.fold_left (fun l1 x -> if p x then l1 else x::l1) [] l);;

let enlever x l = 
  (* suppression de l'lment x de la liste l *)
  suppression (fun y -> x = y) l;;

let rec ens_de_liste l = 
  (* ens_de_liste supprime les lments rpts de l *)
  match l with
    | [] -> []
    | t::f -> if List.mem t f then ens_de_liste f else t::(ens_de_liste f);;


let red infe x l1 =
  (* red infe  x l1 = (a, l2) o
   a vaut true si et seulement si l1 comporte un lment strictement infrieur  x
   l2 est la liste obtenue en enlevant de l1 les lments suprieurs ou gaux  x *)
  let rec aux = function
    | [] -> (false, [])
    | y::l2 -> let (a,l3) = aux l2 in 
      if (infe x y) then (a, l3) else if (infe y x) then (true,y::l3) else (a,y::l3)
  in aux l1;;
      

let rduction infe l1 =
  (* calcule la liste des lments minimaux de l1 pour l'ordre partiel infe *)
  let rec aux = function
    | [] -> []
    | x::l2 -> let (a, l3) = red infe x l2 in
      if a then aux l3 else x::aux l3
  in aux l1;;


let inclus l1 l2 = 
  (* vrifie que la liste l1 est incluse dans la liste l2 *)
  let rec aux = function
    | [] -> true
    | x::l3 -> (List.mem x l2) && aux l3
  in aux l1;;

let limination_inclusion l  = 
  (* enlve de l les listes non minimales pour l'inclusion entre listes *)
  rduction inclus  l;;

let incluant  l1 l2 =
  (* vrifie que l1 inclut un lment de la liste l2 *)
  (List.exists (fun c -> inclus c l1) l2);;
  
let simplification l1 l2 =
  (* enlve de l1 les listes incluants des listes de l2 *)
  suppression  (fun c -> incluant c l2)  l1;;


(* FIN FONCTIONS GENERALES LISTES *)


(* ======================================================================================================
=======================  RESOLUTION OU CONSENSUS    ====================================================
Les noms choisis font allusions au rsolvant, suite aux habitudes du programmeur, mais
vous pouvez remplacer rsolution par consensus et valide par contradictoire si vous le prfrez 
=======================================================================================================*)

open Types;;


	       
let oppose l1 l2 = 
  (* oppose l1 l2 = true si et seulement si l1 et l2 sont des littraux opposs *)
  match (l1, l2) with
    | (Pos p1, Neg p2) -> p1 = p2
    | (Neg p1, Pos p2) -> p1 = p2
    | _ -> false ;;

let opp = 
  (* opp l = le littral oppos  l *)
  function 
    | (Pos x) -> Neg x
    | (Neg x) -> Pos x
;;

let var = 
  (* var l = si l est un littral positif alors l sinon l'oppos de l *)
  function
    | (Pos p1) as l1 -> l1
    | (Neg p1) -> Pos p1 ;;


let rec valide e = match e with
  (* valide e = true ssi e est une clause non numrote valide *)
   | [] -> false 
   | t::f -> (List.mem (opp t) f) || valide f;;

let valide_clause c = 
  (* une clause numrote est valide si et seulement si son corps l'est *)
  valide c.corps ;;

let suppression_valides  l = 
  (* enlve de l les clauses numerotes valides *)
  suppression valide_clause l;;

let rsolvant l c1 c2 = 
(* l est un litteral, c1 et c2 deux clauses numrotes , si l' le  complmentaire de l
est lment de c2 alors reduire l c1 c2 est la clause numerote suivante
   numero = 0
   pere = c1.numero
   mere = c2.numero
   corps =  c1'@ c2',
   c1' tant obtenue en enlevant l du corps de c1 et c2' en enlevant l' du corps de c2
sinon c'est la liste vide *)
  if (List.mem (opp l) c2.corps) 
  then 
    [{numero=0; pere = c1.numero; mere = c2.numero;
      corps = ens_de_liste ((enlever l c1.corps)@(enlever (opp l) c2.corps))}] else [];;

let rsolvants c1 c2 =
(* produit la liste de tous les rsolvants des clauses numerotes c1 et c2 *)
  List.fold_left (fun lc l -> lc@(rsolvant l c1 c2) ) [] c1.corps;;

let rsolvants_clause_liste_de_clauses c lc =  
  (* produit la liste de tous les rsolvants de la clause numerote c et la liste de clauses 
     numerotes lc *)
  List.fold_left (fun lr d -> lr@(rsolvants c d)) [] lc;;

let rsolvants_liste_de_clauses_liste_de_clauses lc1 lc2 =
  (* produit la liste de tous les rsolvants entre les listes de clauses numerotes lc1 et lc2 *)
  List.fold_left (fun lr c -> lr@(rsolvants_clause_liste_de_clauses c lc2)) [] lc1;;

let rec rsolvants_liste_de_clauses lc = 
  (* produit la liste de tous les rsolvants entre les clauses de lc *)
  match lc with
    | [] -> []
    | t::f -> (rsolvants_clause_liste_de_clauses t f)@(rsolvants_liste_de_clauses f);;


let inclus_clauses c1 c2 = 
  (* vrifie que la clause numerote c1 est incluse dans la clause numrote  c2 *)
  let l1 = c1.corps and l2 = c2.corps in
  let rec aux = function
    | [] -> true
    | x::l3 -> (List.mem x l2) && aux l3
  in aux l1;;

let limination_inclusion_clauses l  = 
  (* enlve de l les clauses numrotes  non minimales pour l'inclusion entre clauses numerotes *)
  rduction inclus_clauses  l;;

let incluant_clauses  c l =
  (* vrifie qu'une clause numerote de la liste l est incluse dans la clause numrote c *)
  (List.exists (fun d -> inclus_clauses d c) l);;
  
let simplification_clauses l1 l2 =
  (* enlve de l1 les clauses incluants des clauses de l2 *)
  suppression  (fun c -> incluant_clauses c l2)  l1

let numc = ref 0;;

let rec numeroter l1  = 
  (* numerote les clauses de l1  partir de !numc +1 et modifie numc *)
  match l1 with
    | [] -> []
    | t::f -> numc := !numc +1 ; 
	let a = !numc in {t with numero = a}::numeroter f;;

let print_littral l  = 
  (* impression du littral l *)
  match l with
    | Pos v -> print_string v ; 
    | Neg v -> print_string "-"; print_string v;
;;


let print_litteraux ll =
  print_string "["; 
  if ll <> [] then
    (print_littral (List.hd ll); 
     List.iter (fun l -> print_string " "; print_littral l) (List.tl ll));
  print_string "]";;

let print_preuve t c = 
  (* t est un tableau contenant toutes les preuves, on en extrait la preuve de c 
   chaque clause affiche est marque pour viter les rptitions de preuve *)
  let m = Array.create (Array.length t) false in 
  let rec aux  c = 
    m.(c.numero) <- true;
    if c.pere = 0 then
      (print_int c.numero; print_string " "; print_litteraux c.corps; print_string "\n")
    else
      ( if not m.(c.pere) then aux t.(c.pere);
	if not m.(c.mere) then aux t.(c.mere);
	(* affichage de c (fin de preuve) *)
	print_int c.numero; print_string " "; print_litteraux c.corps; 
	print_string " de "; print_int c.pere; print_string " et "; print_int c.mere ; print_string "\n")
  in aux c;;
   
  
let print_preuves f tce =
  (* affiche les preuves des clauses de f
     tce est la liste des clauses necessaires pour retrouver les preuves
     numc est une rfrence sur le nombre de clauses de tce
     *)
  if f <> [] then
    (let t = Array.create (!numc+1) {numero=0;pere=0;mere=0;corps=[]} in
       (* dans t.(i) on met la clause de numero i*)
       List.iter (fun c -> t.(c.numero)<- c) tce;
       (* on afficher les preuves des lements de f *)
       print_string("Preuves obtenues :\n\n");
       List.iter (fun c -> print_preuve t c; print_string "\n\n") f);;

let print_clause_sans_numero  c =
  (* on affiche une clause sans numero *)
  print_litteraux c.corps; 
  if c.pere =  0
  then  print_string "\n"
  else (print_string " de "; print_int c.pere; print_string " et "; print_int c.mere ; print_string "\n");;

let print_clause c =
  (* on affiche une clause avec numero *)
  print_int c.numero; print_string " "; 
  print_clause_sans_numero c;;

let rec sc n e f tce = 
  (* l'algorithme 31 pour calculer les suites e(i), f(i) pour i > n
     Prconditions : e = e(n), f = e(n), tce est la liste des clauses des ensembles e(i), f(i) pour 
     i au plus gal  n.
     La fonction sc affiche les preuves de toutes les clauses figurant dans son rsultat
  *)
  if e = [] 
  then (print_string "L'ensemble e("; print_int n; print_string") est vide\n\n";
	if f = [] 
	then (print_string "L'ensemble f("; print_int n; print_string") est vide\n\n")
	else
	  ( print_string "L'ensemble f("; print_int n; print_string") comporte les clauses  :\n";
	    List.iter print_clause f; print_string "\n");
	print_preuves f tce ; f) 
 else 
    (* trace des ensembles e(n) et f(n) *)
    (if e = [] then (print_string "L'ensemble e("; print_int n; print_string") est vide\n\n")
     else (print_string "l'ensemble e("; print_int n; print_string ") comprend les clauses :\n";
	   List.iter print_clause e; print_string "\n");
     if f = [] then (print_string "L'ensemble f("; print_int n; print_string") est vide\n\n")
     else (print_string "l'ensemble f("; print_int n; print_string ") comprend les clauses :\n";
	   List.iter print_clause f; print_string "\n"); 
     (* fin des traces *)

     let ef = e@f in 
     let lre = rsolvants_liste_de_clauses e in
     let lref = rsolvants_liste_de_clauses_liste_de_clauses e f in
     let lr1 = lre @ lref in
       ( (* trace des rsolvants calcules *)
	 if lr1 <> [] then 
	   (print_string "resolvants calculs :\n";
	    List.iter print_clause_sans_numero lr1; print_string "\n");
	 let lr2 = suppression_valides lr1 in
	 let lr3 = limination_inclusion_clauses lr2 in
	 let lr4 = simplification_clauses lr3 ef in
	 let en = numeroter lr4 in
	 let fn  = simplification_clauses  ef lr4 in
	 (* let fn = numeroter lr5 in *)
	   sc (n+1) en fn  (en@fn@tce)));;

let algo31 e =
  let e1 = suppression_valides e in
  let e2 = limination_inclusion_clauses e1 in 
  let e3 = numeroter e2 in
    sc 0 e3 [] e3;;





(*  let e1 = numeroter e in
  let e2 = suppression_valides e1 in
  let e3 = limination_inclusion_clauses e2 in sc 0 e3 [] e2;;*)

(* FIN STATEGIE COMPLETE DANS LA VERSION AVEC TRACE *)


(* =========================================================================================
=========================== PETITE INTERFACE ===============================================
===========================================================================================*)


let afficher_aide () =
  print_string "une commande doit tre suivie par ;;\n"; 
  print_string "les commandes sont:\n";
  print_string "\tquitter\n";
  print_string "\taide\n"; 
  print_string "\tpremier suivi d'une liste de clauses\n";
  print_string "\t\tles clauses sont spares par des virgules\n";
  print_string "\t\tune clause est une suite de littraux\n";
  print_string "\t\tun littral positif est un identificateur\n";
  print_string "\t\tun littral ngatif est un identificateur prcd du signe moins\n";;



exception Bye;;

let executer_commande c = 
  (* excute la commande c *)
  match c with
    | Quitter -> raise Bye
    | Aide -> afficher_aide (); print_newline ()
    | Premier l -> numc:= 0; let _ = algo31 l in ()
;;

let prompt = ">>";;

let lire_commande channel = Parser.commande Lexer.terminal (Lexing.from_channel channel);;

let interaction () =
  (* la boucle d'interaction traditionnelle *)
  try while true do
    try print_string prompt; flush stdout; executer_commande (lire_commande stdin)
    with 
      | Lexer.Erreur_lexicale -> print_string "Erreur lexicale\n"
      | Parsing.Parse_error -> print_string "Erreur syntaxique\n"
  done
  with Bye -> ();;

print_string "Taper \"aide;;\" pour obtenir des informations";
print_newline ();
interaction () ;;

 
(* ==========================================================================================
==================  EXEMPLE DE P.AMBLARD ====================================================
============================================================================================

   -x3 -x2 x1 x0, -x3 -x2 x1 -x0, -x3 x2 -x1 -x0, -x3 x2 -x1 x0,
   -x3 x2 x1 -x0, x3 x2 -x1 x0, x3 x2 x1 x0, x3 x2 x1 -x0, 
   x3 -x2 -x1 -x0, x3 -x2 -x1 x0, x3 -x2 x1 x0, x3 -x2 x1 -x0

Comme peut le constater Paul Amblard, cet algorithme calcule les monmes premiers de la fonction
XXX dans l'ordre
e = x2'x1
a = x1x0'
c = x3x0
b = x3x1
d = x3x2'
g = x3'x2x1'
h = x3'x2x0'
f = x2x1'x0
*)
